package main

import (
	"bufio"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"path/filepath"
	"strconv"
	"strings"

	"github.com/orjansv/master-code/utils"

	"github.com/orjansv/master-code/benchmark"
	"github.com/orjansv/master-code/simulator"
	"github.com/orjansv/master-code/simulator/cloud"
)

var numMethods = 1
var numGrids = 1

func main() {

	// Get the settings from the arguments
	args := os.Args[1:]
	settings := handleArgs(args)
	if settings.timesToSimulate == 0 {
		settings.timesToSimulate = 1
	}

	// Initialize variables
	allBenchmarks := make(map[string]*benchmark.Benchmark)
	for _, types := range cloud.SelectionMethodsString {
		allBenchmarks[types] = &benchmark.Benchmark{}
	}

	gridsJSON := GridsJSON{}
	lastGrid := -1
	allMessages := MessagesJSON{}
	fullResults := []SimulationEnergyInfo{}

	b := benchmark.GetBenchmarkStats(settings.filename)
	if settings.benchmark {
		numMethods = 4
	}

	if b.UsesMultiCoordinates && settings.benchmark {
		numGrids = b.NumCoordinates
	}

	// Intro. Wait for user to start.
	waitForUser()

	// Simulate on all methods if benchmarking. Otherwise, only the selected method.
	for method := 0; method < numMethods; method++ {
		for grid := 0; grid < numGrids; grid++ {
			for i := 0; i < settings.timesToSimulate; i++ {

				printSimulationInfo(i, grid)

				// Load configurations from file
				simulator := simulator.Simulator{VerboseLevel: settings.verbose}
				simulator.ParseConfigFile(settings.filename)

				// Adjust simulator settings according to benchmark.
				if settings.benchmark {
					// We want to benchmark with all selection methods
					simulator.SetSelectionMethod(method)

					if b.UsesMultiCoordinates {
						simulator.SetPositions(b.SensorCoordinates[grid], b.GatewayCoordinates[grid])
					}
				}

				// We add grid + i to the seed. Method is left out to make sure all methods are compared equally.
				simulator.Setup(grid + i)

				// Run simulation.
				results, err := simulator.Run()
				if err != nil {
					panic(err)
				}

				// This check is needed because we only plot each grid once.
				if grid != lastGrid && method == 0 {
					results.Connections.Number = grid + 1
					gridsJSON.Grids = append(gridsJSON.Grids, results.Connections)
					lastGrid = grid
				}

				// Store messages for writing (if enabled through options).
				if settings.saveMessages {
					msgSet := MessageSet{
						Messages:  results.Messages,
						Method:    cloud.SelectionMethodsString[method],
						Grid:      grid + 1,
						Iteration: i + 1,
					}
					allMessages.MessageSets = append(allMessages.MessageSets, msgSet)
				}

				// Save benchmark settings and increment.
				if settings.benchmark {
					m := cloud.SelectionMethodsString[method]
					allBenchmarks[m].Retries = append(allBenchmarks[m].Retries, results.Performance.TotalRetries)
					allBenchmarks[m].TotalRetries += results.Performance.TotalRetries
					allBenchmarks[m].Energy = append(allBenchmarks[m].Energy, results.Performance.TotalEnergyConsumption)
					allBenchmarks[m].TotalEnergyConsumption += results.Performance.TotalEnergyConsumption
					allBenchmarks[m].Boosts = append(allBenchmarks[m].Boosts, results.Performance.TotalBoostMessages)
					allBenchmarks[m].TotalBoost += results.Performance.TotalBoostMessages

					res := SimulationEnergyInfo{
						Grid:       grid + 1,
						Iteration:  i + 1,
						Method:     cloud.SelectionMethodsString[method],
						EnergyInfo: results.Performance,
					}
					fullResults = append(fullResults, res)
				}

			}
		}
	}

	relPath, name := getPathAndNameFromFile(b.Name)
	newPath := getStoragePath(relPath)

	title := name
	if len(b.Title) > 0 {
		title = b.Title
	}

	// Save json file for plotting grids.
	gridsJSON.Title = title
	savename := newPath + "\\" + name + "_connections.json"
	writeStructToFile(gridsJSON, savename)

	// Write messages to message file.
	if settings.saveMessages {
		savename = newPath + "\\" + name + "_messages.json"
		writeStructToFile(allMessages, savename)
	}

	// Write benchmark results to files.
	if settings.benchmark {
		benchmarkJSON := BenchmarkJSON{}
		for _, m := range cloud.SelectionMethodsString {
			allBenchmarks[m].Method = strings.Replace(m, "_", " ", 3)
			benchmarkJSON.Benchmarks = append(benchmarkJSON.Benchmarks, *allBenchmarks[m])
		}

		benchmarkJSON.Title = b.Title

		// Write JSON to file.
		savename = newPath + "\\" + name + "_benchmark.json"
		writeStructToFile(benchmarkJSON, savename)

		// Save all results
		savename = newPath + "\\" + name + "_all_results.json"
		writeStructToFile(fullResults, savename)
	}

	// Plot grid with sensors, gateways and their connections.
	if settings.plot == true {

		fmt.Printf("\nSimulation stopped. Plotting results.\n")

		if settings.benchmark {
			// Plot simulation results with python script.
			err := exec.Command("python", "stats.py", newPath+"/"+name+"_benchmark.json").Run()
			if err != nil {
				fmt.Println(err)
			}
		}

		// Plot grid(s)
		err := exec.Command("python", "plot.py", newPath+"/"+name+"_connections.json", "-sq").Run()
		if err != nil {
			fmt.Println(err)
		}

		fmt.Println("Finished plotting.")
	} else {
		fmt.Printf("\nSimulation stopped.\n")
	}

}

func waitForUser() {
	reader := bufio.NewReader(os.Stdin)
	fmt.Println("")
	fmt.Println("Welcome to this WSN simulator!")
	fmt.Println("Press return to start the execution. When running, press return to stop.")
	_, _ = reader.ReadString('\n')
	return
}

func printSimulationInfo(i, grid int) {
	if numGrids > 1 {
		fmt.Printf("\nSIMULATION %2d - GRID %2d\n", i+1, grid+1)
		fmt.Printf("-----------------------\n")
	} else {
		fmt.Printf("\nSIMULATION %2d\n", i+1)
		fmt.Printf("-------------\n")
	}
}

func writeStructToFile(strct interface{}, filename string) {

	bytes, err := json.Marshal(strct)
	if err != nil {
		panic(err)
	}

	err = ioutil.WriteFile(filename, bytes, 0644)
	if err != nil {
		panic(err)
	}
}

func getPathAndNameFromFile(filepath string) (string, string) {

	// Check name for slash
	path := ""
	temp := strings.Split(filepath, "\\")
	name := filepath
	if len(temp) > 1 {
		path = filepath
		name = temp[len(temp)-1]
	}

	return path, name
}

func getStoragePath(dir string) string {

	// Create directory dir in current path
	ex, err := os.Executable()
	if err != nil {
		panic(err)
	}
	exPath := filepath.Dir(ex)
	newPath := exPath + "/" + dir

	if _, err := os.Stat(newPath); os.IsNotExist(err) {
		err = os.Mkdir(newPath, 0644)
		if err != nil {
			newPath = "" // simply use current folder
		}
	}
	if err != nil {
		newPath = "" // simply use current folder
	}

	return newPath
}

func handleArgs(args []string) runSettings {

	r := runSettings{}
	other := []string{}
	num := false

	if len(args) == 0 {
		printHelp()
		os.Exit(1)
	}

	for _, arg := range args {

		if arg[0] == '-' {
			switch arg {

			case "-b":
				r.benchmark = true
				r.plot = true

			case "-h":
				printHelp()
				os.Exit(1)

			case "-j":
				createEmptyJSON()

			case "-m":
				r.saveMessages = true

			case "-n":
				num = true

			case "-p":
				r.plot = true

			case "-v":
				r.verbose = 2

			case "-v1":
				r.verbose = 1

			case "-v2":
				r.verbose = 2

			default:
				fmt.Printf("simulate: unknown option %s\n", arg)
				fmt.Printf("Try 'simulate -h' for more information.\n")
				os.Exit(1)
			}
		} else {
			if num == true {
				times, err := strconv.Atoi(arg)
				if err != nil {
					panic(err)
				}
				r.timesToSimulate = times
				num = false
			} else {
				other = append(other, arg)
			}
		}
	}

	if len(other) > 1 {
		err := fmt.Sprintf("Too many arguments provided: %v. Please only provide one .json file with configurations.\n", other)
		panic(err)
	} else {
		var extension = filepath.Ext(other[0])
		if extension != ".json" {
			panic("The configuration file must be of type json.")
		}
		r.filename = other[0]
	}

	return r
}

func printHelp() {
	fmt.Printf("Wireless Sensor Simulator 1.0\n")
	fmt.Printf("Usage: simulate [FILE] [OPTION]...\n")
	fmt.Printf("Simulates a wireless sensor network.\n\n")
	fmt.Printf("The file needs to be a configuration file in JSON format.\n\n")
	fmt.Println("Available options:")
	fmt.Printf("  -b  benchmark, can be combined with -n\n")
	//fmt.Printf("  -e gives an example of a minimal JSON configuration file.\n")
	fmt.Printf("  -h  display this help and exit\n")
	fmt.Printf("  -j  creates an empty json file with all settings named empty.json\n")
	fmt.Printf("  -m  save all messages in JSON file\n")
	fmt.Printf("  -n  number of times to run the simulation\n")
	fmt.Printf("  -p  plot grid with sensors and gateways\n")
	fmt.Printf("  -v  verbose mode, also comes with levels: -v1: print commands, -v2: print commands and messages\n")
}

func createEmptyJSON() {
	empty := simulator.ConfigJSON{}
	empty.Sensors.Coordinates = make([]utils.Coordinates, 2)
	empty.Sensors.MultiCoordinates = make([][]utils.Coordinates, 2)
	for i := range empty.Sensors.MultiCoordinates {
		empty.Sensors.MultiCoordinates[i] = make([]utils.Coordinates, 2)
	}
	empty.Gateways.Coordinates = empty.Sensors.Coordinates
	empty.Gateways.MultiCoordinates = empty.Sensors.MultiCoordinates
	writeStructToFile(empty, "empty.json")
	fmt.Println("Successfully created empty.json.")
	os.Exit(0)
}
