import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
import math
import json
import os.path
import sys
from matplotlib import rc

if len(sys.argv) > 3:
  print("Error: This program takes one or two arguments: name of json file and optionally -s for subplot.")
  exit(1)

if sys.argv[1].endswith('json') != True:
  print("Error: The file has to be a .json file.")
  exit(1)

subplot = False
square = False
if len(sys.argv) == 3:
  if sys.argv[2] == '-s':
    subplot = True
  elif sys.argv[2] == '-sq':
    subplot = True
    square = True # special square grid for thesis

# Name without extension and "_stats".
tmp = os.path.splitext(sys.argv[1])[0]
name = tmp.split('_')[0]

json_data=open(sys.argv[1]).read()
allData = json.loads(json_data)

connectioncolors = {
  'standard': '#abdda4',
  'dynamic': '#2b83ba',
  'binary': '#d7191c',
  'boost': '#fdae61'
}

colors = ['#2b83ba','#fdae61','#abdda4','#d7191c']

connectionstyle = {
  'standard': '-',
  'dynamic': '--',
  'binary': ':',
  'boost': '-.'
}

# LaTeX fonts :-)
rc('font', **{'family': 'serif', 'serif': ['Computer Modern']})
rc('text', usetex=True)

numGrids = len(allData["Grids"])
for k, data in enumerate(allData["Grids"]):

  plt.clf()
  x_max = 0
  y_max = 0

  # Set size in inches (this might not be needed)
  width=8.0
  height=6.0

  sensors = []
  for i, val in enumerate(data["sensors"]):
    sensors.append((val["x"], val["y"]))
    if val["x"] > x_max:
      x_max = val["x"]
    if val["y"] > y_max:
      y_max = val["y"]

  gateways = []
  for i, val in enumerate(data["gateways"]):
    gateways.append((val["x"], val["y"]))
    if val["x"] > x_max:
      x_max = val["x"]
    if val["y"] > y_max:
      y_max = val["y"]


  fig, ax = plt.subplots()

  if numGrids > 1:
    plt.title(allData["Title"].title() + " - Grid " + str(data["number"]))
  else:
    plt.title(allData["Title"].title())

  ax.set_xlim([-0.5, x_max + 0.5])
  ax.set_ylim([-0.5, y_max + 0.5])
  ax.set_aspect('equal')
  plt.grid()

  # 72 points per inch
  point_height=height*72

  x1,x2,y1,y2 = plt.axis()
  yrange = y2 - y1

  lineWidth = (0.2 * (point_height/yrange)) * 0.8 # corresponding in points?
  if lineWidth > 2:
    lineWidth = 2
  if lineWidth < 0:
    lineWidth = 1

  sensorSize = (0.25 * (point_height/yrange)) * 0.8 #sensors should be half a unit
  if sensorSize > 8:
    sensorSize = 8
  if sensorSize < 2:
    sensorSize = 2

  gatewaySize = (0.5 * (point_height/yrange)) * 0.8 #gateways should be one full unit
  if gatewaySize > 16:
    gatewaySize = 16
  if gatewaySize < 4:
    gatewaySize = 4

  if subplot != True:
    for i, type in enumerate(data["connections"]):
      if type != "boost":
        if data["connections"][type] is not None:
          for j, val in enumerate(data["connections"][type]):
            ax.plot(val["x"], val["y"], color=connectioncolors[type], linestyle=connectionstyle[type], alpha=0.7, linewidth=lineWidth, label=type.title() if j == 0 else "")


    ax.plot(*zip(*sensors), marker='o', linestyle='none', label="Sensor", color='black', markersize=sensorSize, markeredgecolor='black')
    ax.plot(*zip(*gateways), marker='s', linestyle='none', label="Gateway", color='black', markersize=gatewaySize, markeredgecolor='black')

    if (x_max < 15.0) and (y_max < 15.0):
      plt.legend(labelspacing=1.5, borderpad=1, fancybox=False, edgecolor='black')
    else:
      plt.legend(fancybox=False, edgecolor='black')

    plt.savefig(name + '_grid_' + str(data["number"]) + '.pdf', bbox_inches='tight')
    plt.close()

    # Only boost
    if data["connections"]["boost"] is not None:
      fig, ax = plt.subplots()
      for i, type in enumerate(data["connections"]):
        if data["connections"][type] is not None:
          for j, val in enumerate(data["connections"][type]):
            ax.plot(val["x"], val["y"], color=connectioncolors["boost"], linestyle=connectionstyle["boost"], alpha=0.7, linewidth=lineWidth, label=type.title() if type == "boost" and j == 0 else "")

      ax.plot(*zip(*sensors), marker='o', linestyle='none', label="Sensor", color='black', markersize=sensorSize, markeredgecolor='black')
      ax.plot(*zip(*gateways), marker='s', linestyle='none', label="Gateway", color='black', markersize=gatewaySize, markeredgecolor='black')

      if (x_max < 15.0) and (y_max < 15.0):
        plt.legend(labelspacing=1.5, borderpad=1, fancybox=False, edgecolor='black')
      else:
        plt.legend(fancybox=False, edgecolor='black')

      ax.set_xlim([-0.5, x_max + 0.5])
      ax.set_ylim([-0.5, y_max + 0.5])
      ax.set_aspect('equal')
      plt.grid()

      if numGrids > 1:
        plt.title(allData["Title"].title() + " - Grid " + str(data["number"]) + " - Boost")
      else:
        plt.title(allData["Title"].title() + " - Boost")

      plt.savefig(name + '_grid_' + str(data["number"]) + '_boost.pdf', bbox_inches='tight')
      plt.close()

  else:

    # Both in separate sub-plots
    if data["connections"]["boost"] is not None:
      fig, (ax1, ax2) = plt.subplots(1,2)

      #ax1.set_xlim([-1.0, x_max + 1.0])
      #ax1.set_ylim([-1.0, y_max + 1.0])

      #ax2.set_xlim([-1.0, x_max + 1.0])
      #ax2.set_ylim([-1.0, y_max + 1.0])

      ax1.set_xlim([-1.0, 50 + 1.0])
      ax1.set_ylim([-1.0, 50 + 1.0])

      ax2.set_xlim([-1.0, 50 + 1.0])
      ax2.set_ylim([-1.0, 50 + 1.0])



      for i, type in enumerate(data["connections"]):
          if type != "boost":
            if data["connections"][type] is not None:
              for j, val in enumerate(data["connections"][type]):
                ax1.plot(val["x"], val["y"], color=connectioncolors[type], linestyle=connectionstyle[type], alpha=0.7, linewidth=lineWidth, label=type.title() if j == 0 else "")

      for j, val in enumerate(data["connections"]["boost"]):
        ax2.plot(val["x"], val["y"], color=connectioncolors["boost"], linestyle=connectionstyle["boost"], alpha=0.7, linewidth=lineWidth, label="Boost" if j == 0 else "")

      ax1.plot(*zip(*sensors), marker='o', linestyle='none', color='black', markersize=sensorSize, markeredgecolor='black')
      ax1.plot(*zip(*gateways), marker='s', linestyle='none', color='black', markersize=gatewaySize, markeredgecolor='black')
      ax2.plot(*zip(*sensors), marker='o', linestyle='none', label="Sensor", color='black', markersize=sensorSize, markeredgecolor='black')
      ax2.plot(*zip(*gateways), marker='s', linestyle='none', label="Gateway", color='black', markersize=gatewaySize, markeredgecolor='black')

      if (x_max < 15.0) and (y_max < 15.0):
        lgd = fig.legend(ncol=3, loc=(0.1875, 0.05), shadow=True, fancybox=False, edgecolor='black', labelspacing=1.5, borderpad=1,)
      else:
        lgd = fig.legend(ncol=3, loc=(0.1875, 0.05), shadow=True, fancybox=False, edgecolor='black')

      # Hard coded for thesis
      if square == True:
        loc = plticker.MultipleLocator(base=10) # this locator puts ticks at regular intervals
        ax1.xaxis.set_major_locator(loc)
        ax1.yaxis.set_major_locator(loc)
        ax2.xaxis.set_major_locator(loc)
        ax2.yaxis.set_major_locator(loc)

      ax1.grid(True)
      ax1.set_aspect('equal')
      ax2.grid(True)
      ax2.set_aspect('equal')

      if numGrids > 1:
        ax1.set_title(allData["Title"].title() + " - Grid " + str(data["number"]) + " - Standard")
        ax2.set_title(allData["Title"].title() + " - Grid " + str(data["number"]) + " - Boost")
      else:
        ax1.set_title(allData["Title"].title() + " - Standard")
        ax2.set_title(allData["Title"].title() + " - Boost")

      plt.savefig(name + '_grid_' + str(data["number"]) + '_both.pdf', bbox_inches='tight')
      plt.close()
