package cloud

import (
	"github.com/orjansv/master-code/simulator/sensor"
)

// HandleMessage handles incoming messages from the gateways.
func (c *Cloud) HandleMessage(msg sensor.Message) {
	connKey := ConnectionKey{msg.SensorID, msg.GatewayID}
	c.checkInitialization(connKey, msg)

	// Keep track of available replies at responsible gateway. TODO: MOVE?
	if msg.GatewayID == c.responsibleGateway[msg.SensorID].ID {
		if c.responsibleGateway[msg.SensorID].RepliesAvailable > 0 {
			c.responsibleGateway[msg.SensorID].RepliesAvailable--
		}
	}

	// Only change gateway if we received a new message (never on retries)
	msgKey := MessageKey{msg.SensorID, msg.MessageID}
	_, oldMessage := c.messages[msgKey] // Only change gateway if we receive a new
	newMessage := !oldMessage

	if newMessage {
		// The first gateway to forward a new message is automatically the next optimal candidate.
		c.responsibleGateway[msg.SensorID].NextResponsible = msg.GatewayID

		// Move retry window one step for new messages
		c.retries[connKey].Move(msg.MessageID)

		go c.startChangeGwTimer(msg.SensorID)
	}

	responsibleKey := ConnectionKey{msg.SensorID, c.responsibleGateway[msg.SensorID].ID}

	// Select gateway based on selection method
	switch c.selectionMethod {
	case SelectionMethodStrongestSignal:
		c.updateAverageSignalStrength(connKey, msg)
		c.selectGatewaySignalStrength(msg, connKey)

	case SelectionMethodLowestRetryCount:
		c.updateAverageSignalStrength(connKey, msg)
		c.countRetries(responsibleKey, msg)
		c.selectGatewayLowestRetryCount(msg, connKey)

	case SelectionMethodHighestReliability:
		c.monitorUplink(msg)
		c.monitorDownlink(msg, connKey, responsibleKey)
		c.selectGatewayReliability(msg, connKey)

	case SelectionMethodLowestEnergyConsumption:
		c.monitorUplink(msg)
		c.monitorDownlink(msg, connKey, responsibleKey)
		c.updateExpectedEnergyConsumption(msg, connKey)
		c.selectGatewayEnergy(msg, connKey)
	}

	// Make sure optimal gateway has enough replies available. TODO: hard coded 4? Fix?
	c.SendReplies(msg.SensorID, 4-c.responsibleGateway[msg.SensorID].RepliesAvailable)
}

// PrintMessage prints a message received by the cloud.
func (c *Cloud) PrintMessage(msg sensor.Message, timestamp string) {

	if c.verboseLevel > 1 {
		switch c.selectionMethod {
		case SelectionMethodStrongestSignal:
			k := ConnectionKey{msg.SensorID, msg.GatewayID}
			printMessageSignalStrength(msg, timestamp, c.connections[k].SignalStrength.Average())

		case SelectionMethodLowestRetryCount:
			k := ConnectionKey{msg.SensorID, msg.GatewayID}
			printMessageRetryCount(msg, timestamp, c.retries[k].TotalCount())

		case SelectionMethodHighestReliability:
			k := ConnectionKey{msg.SensorID, msg.GatewayID}
			wRel := c.calculateWeightedReliability(k, c.sensor.BoostModeEnabled)
			printMessageReliability(msg, timestamp, wRel)

		case SelectionMethodLowestEnergyConsumption:
			k := ConnectionKey{msg.SensorID, msg.GatewayID}
			ex := c.connections[k].ExpectedEnergyConsumption.Average()

			pu, pd, puBoost, pdBoost := -1.0, -1.0, -1.0, -1.0
			if c.upLink.Available[k] {
				pu = c.upLink.Reliability(k)
				pd = c.downlinks[k].Reliability()
			}

			if c.upLinkBoost.Available[k] {
				puBoost = c.upLinkBoost.Reliability(k)
				pdBoost = c.downlinksBoost[k].Reliability()
			}
			printMessageEnergy(msg, timestamp, ex, pu, pd, puBoost, pdBoost)
		}
	}
}

// InsertMessage inserts and keeps track of all incoming messages as events.
func (c *Cloud) InsertMessage(msg sensor.Message, timestamp string) {

	key := MessageKey{msg.SensorID, msg.MessageID}
	connectionInfo := ConnectionInfo{msg.GatewayID, msg.SignalStrength, timestamp}

	// <SensorID, MessageID> event has already happened at least once.
	if _, ok := c.messages[key]; ok {

		c.messages[key].Ciphertext[msg.Ciphertext] = true
		c.messages[key].Gateways = append(c.messages[key].Gateways, connectionInfo)

	} else {
		// <SensorID, MessageID> event has happened for the first time.
		ciphertext := map[string]bool{msg.Ciphertext: true}
		c.messages[key] = &Event{msg.SensorID, ciphertext, []ConnectionInfo{connectionInfo}}
	}

}

// MessageIsRetry checks whether a message was a retry.
func (c *Cloud) MessageIsRetry(msg sensor.Message) bool {

	key := MessageKey{msg.SensorID, msg.MessageID}
	if event, ok := c.messages[key]; ok {
		if event.Ciphertext[msg.Ciphertext] == true {
			// Ciphertext has already been seen, so if it was a retry, it has already been counted.
			return false
		}

		// Ciphertext has not been seen before, but we must make sure it is not actually the first ciphertext for this ID.
		// If there has been no ciphertext for this event before, then this is the first one, so not a retry.
		// Otherwise this is a new ciphertext for an event and hence it's a retry.
		if len(event.Ciphertext) == 0 {
			return false
		}
		return true
	}

	return false
}

// MessageIsNew checks whether this message has been received before or not.
func (c *Cloud) MessageIsNew(msg sensor.Message) bool {
	key := MessageKey{msg.SensorID, msg.MessageID}

	if event, ok := c.messages[key]; ok {
		// Message ID has previously been received.
		// What about the ciphertext?
		if event.Ciphertext[msg.Ciphertext] == true {
			// Ciphertext has already been seen, so not a new message.
			return false
		}

		// New ciphertext, so count it as a new message.
		return true
	}

	// This message ID has not been previously received, so it should be counted.
	return true
}

func (c *Cloud) countRetries(key ConnectionKey, msg sensor.Message) {
	if c.MessageIsRetry(msg) {
		c.retries[key].Insert(msg.MessageID)
	}
}
