package connection

import (
	"math"
	"math/rand"
)

// Connection is a type used to calculate signal strength for a given connection.
type Connection interface {
	GetSignalStrengthAtSensor(msgID int) float64
	GetSignalStrengthAtGateway(msgID int, transmissionPower float64) float64
}

const (
	// TypeDefault is a standard connection with no obsturctions.
	TypeDefault int = iota

	// TypeDynamic is a door connection that reduces the signal strength when open (Which is random).
	TypeDynamic

	// TypeBinary is a door connection that completely blocks when open (which is random).
	TypeBinary
)

// ConnectionTypes is used for converting json string to number.
var ConnectionTypes = map[string]int{
	"standard": TypeDefault,
	"dynamic":  TypeDynamic,
	"binary":   TypeBinary,
}

// NewConnection creates a new instance of a Connection with a specific type.
func NewConnection(dist, frequency, gatewayTransmissionPower, blockThreshold, adjustment float64, connectionType, blockProb, dynLoss int, seed int64, symmetric bool) Connection {

	var c Connection
	s := rand.NewSource(seed)
	r := rand.New(s)

	// Adjustment is to adjust FSPL according to max range
	freeSpacePathLoss := 20*math.Log10(dist*adjustment) + 20*math.Log10(frequency) - 147.55 // with meters and Hz as units

	dynamicLoss := float64(dynLoss) / 10.0

	switch connectionType {
	case TypeDefault:
		c = &StandardConnection{gatewayTransmissionPower: gatewayTransmissionPower, freeSpacePathLoss: freeSpacePathLoss}
	case TypeDynamic:
		c = &DynamicConnection{
			msgID: -1,
			gatewayTransmissionPower: gatewayTransmissionPower,
			freeSpacePathLoss:        freeSpacePathLoss,
			rnd:                      r,
			symmetric:                symmetric,
			dynamicLoss:              dynamicLoss,
		}
	case TypeBinary:
		c = &BinaryConnection{
			msgID: -1,
			gatewayTransmissionPower: gatewayTransmissionPower,
			freeSpacePathLoss:        freeSpacePathLoss,
			rnd:                      r,
			blockProbability:         blockProb,
			blockThreshold:           blockThreshold - 1,
			symmetric:                symmetric,
		}
	default:
		c = &StandardConnection{gatewayTransmissionPower: gatewayTransmissionPower, freeSpacePathLoss: freeSpacePathLoss}
	}

	return c
}

// StandardConnection is a connection type with only free path loss.
type StandardConnection struct {
	freeSpacePathLoss        float64
	gatewayTransmissionPower float64
}

// GetSignalStrengthAtSensor returns the signal strength of a message received at a sensor from a gateway.
func (c *StandardConnection) GetSignalStrengthAtSensor(msgID int) float64 {
	return c.gatewayTransmissionPower - c.freeSpacePathLoss
}

// GetSignalStrengthAtGateway returns the signal strength at the gateway, depending on the sensor transmission power and the FSPL.
func (c *StandardConnection) GetSignalStrengthAtGateway(msgID int, sensorTransmissionPower float64) float64 {
	return sensorTransmissionPower - c.freeSpacePathLoss
}

// DynamicConnection is a connection type that may be open and thus reduce the signal strength even more than free space.
type DynamicConnection struct {
	msgID                    int
	gatewayTransmissionPower float64
	freeSpacePathLoss        float64
	increasedLoss            float64
	rnd                      *rand.Rand
	symmetric                bool
	dynamicLoss              float64
}

// GetSignalStrengthAtSensor returns the signal strength of a message received at a sensor from a gateway.
func (c *DynamicConnection) GetSignalStrengthAtSensor(msgID int) float64 {

	if c.symmetric {
		if msgID != c.msgID {
			c.increasedLoss = randomIncrease(c.rnd, c.dynamicLoss)
			c.msgID = msgID
		}
	} else {
		c.increasedLoss = randomIncrease(c.rnd, c.dynamicLoss)
	}

	return c.gatewayTransmissionPower - c.freeSpacePathLoss*c.increasedLoss
}

// GetSignalStrengthAtGateway returns the signal strength at the gateway, depending on the sensor transmission power and the FSPL.
func (c *DynamicConnection) GetSignalStrengthAtGateway(msgID int, sensorTransmissionPower float64) float64 {

	if c.symmetric {
		if msgID != c.msgID {
			c.increasedLoss = randomIncrease(c.rnd, c.dynamicLoss)
			c.msgID = msgID
		}
	} else {
		c.increasedLoss = randomIncrease(c.rnd, c.dynamicLoss)
	}

	return sensorTransmissionPower - c.freeSpacePathLoss*c.increasedLoss
}

func randomIncrease(rnd *rand.Rand, percentage float64) float64 {
	return (10.0 - percentage + 2*percentage*rnd.Float64()) / 10.0 // random float between 0.95 and 1.05
}

// BinaryConnection is a connection type that either blocks completely or not at all.
type BinaryConnection struct {
	msgID                    int
	gatewayTransmissionPower float64
	freeSpacePathLoss        float64
	isBlocked                bool
	rnd                      *rand.Rand
	blockProbability         int
	blockThreshold           float64
	symmetric                bool
}

// GetSignalStrengthAtSensor returns the signal strength of a message received at a sensor from a gateway.
func (c *BinaryConnection) GetSignalStrengthAtSensor(msgID int) float64 {

	if c.symmetric {
		if msgID != c.msgID {
			c.isBlocked = (c.rnd.Intn(100) + 1) < c.blockProbability
			c.msgID = msgID
		}
	} else {
		c.isBlocked = (c.rnd.Intn(100) + 1) < c.blockProbability
	}

	if c.isBlocked {
		return c.blockThreshold
	}
	return c.gatewayTransmissionPower - c.freeSpacePathLoss
}

// GetSignalStrengthAtGateway returns the signal strength at the gateway, depending on the sensor transmission power and the FSPL.
func (c *BinaryConnection) GetSignalStrengthAtGateway(msgID int, transmissionPower float64) float64 {

	if c.symmetric {
		if msgID != c.msgID {
			c.isBlocked = (c.rnd.Intn(100) + 1) < c.blockProbability
			c.msgID = msgID
		}
	} else {
		c.isBlocked = (c.rnd.Intn(100) + 1) < c.blockProbability
	}

	if c.isBlocked {
		return c.blockThreshold
	}
	return transmissionPower - c.freeSpacePathLoss
}
