package gateway

import (
	"time"

	"github.com/orjansv/master-code/simulator/common"
	"github.com/orjansv/master-code/simulator/sensor"
	"github.com/orjansv/master-code/simulator/simtime"
)

// Gateway is the type used for simulating a gateway.
type Gateway struct {
	ID          int
	sensitivity float64
	channels    Channels
	replies     map[int]int
}

// Channels is a struct that holds all channels for the gateway.
type Channels struct {
	Gateway2CloudChan  chan sensor.Message
	Cloud2GatewayChan  chan common.CommandMessage
	Gateway2SensorChan map[int]chan sensor.Message
	Sensor2GatewayChan chan sensor.Message
	StopChan           chan int
}

// NewGateway returns a new instance of the gateway type.
func NewGateway(gatewayID int, sensitivity float64, gwChannels Channels) Gateway {

	// Ready replies
	replies := make(map[int]int)

	g := Gateway{gatewayID, sensitivity, gwChannels, replies}
	return g
}

// Run starts the simulation of the gateway.
func (g *Gateway) Run() {

	for {

		// Wait for messages from sensors or the cloud.
		select {
		// Received message from a sensor.
		case msg := <-g.channels.Sensor2GatewayChan:
			if msg.SignalStrength > g.sensitivity {
				// Send reply to sensor if we have any.
				if g.replies[msg.SensorID] > 0 {
					select {
					case g.channels.Gateway2SensorChan[msg.SensorID] <- sensor.Message{SensorID: msg.SensorID, MessageID: msg.MessageID, GatewayID: g.ID, Boost: msg.Boost}:
						// Successfully sent reply to sensor.
						g.replies[msg.SensorID]--

					case <-time.After(1 * simtime.Minute):
						// If the sensor does not accept the answer after 1 minute, we were too late.

					case <-g.channels.StopChan:
						// Confirm stopping
						g.channels.StopChan <- 0
						return
					}
				}

				// Append information to sensor message and forward it to the cloud.
				msg.GatewayID = g.ID

				select {
				case g.channels.Gateway2CloudChan <- msg:
					// Successfully forwarded message to the cloud.
				case <-g.channels.StopChan:
					// Confirm stopping to the simulator.
					g.channels.StopChan <- 0
					return
				}
			} else {

			}

		// Received message from the cloud.
		case msg := <-g.channels.Cloud2GatewayChan:
			switch msg.Command {
			case common.CMDReply:
				g.replies[msg.SensorID] += msg.NumReplies

			case common.CMDFlush:
				g.replies[msg.SensorID] = 0

			}

		case <-g.channels.StopChan:
			// Confirm stopping to the simulator.
			g.channels.StopChan <- 0
			return

		}
	}
}
