package simulator

import (
	"encoding/json"
	"fmt"
	"hash/crc64"
	"io/ioutil"
	"os"
	"sort"
	"time"

	"github.com/orjansv/master-code/simulator/cloud"
	"github.com/orjansv/master-code/simulator/sensor"
	"github.com/orjansv/master-code/simulator/simtime"
)

//ParseConfigFile loads simulation configurations from a configuration file and replaces missing values with defaults.
func (s *Simulator) ParseConfigFile(filename string) {

	// Load configurations from file.
	file, err := ioutil.ReadFile(filename)
	if err != nil {
		panic(err)
	}

	var conf ConfigJSON
	err = json.Unmarshal(file, &conf)
	if err != nil {
		error := fmt.Sprintf("Failed to load config file: %s", err)
		panic(error)
	}

	s.parseGeneralConfig(conf)
	s.parseSensorConfig(conf)
	s.parseGatewayConfig(conf)
	s.parseCloudConfig(conf)
	s.parseConnectionConfig(conf)
}

func (s *Simulator) parseGeneralConfig(conf ConfigJSON) {
	// So far we only have three connection types.
	s.connectionDistribution = make([]float64, numConnectionTypes)
	s.boostConnectionDistribution = make([]float64, numConnectionTypes)
	s.gridJSON = GridJSON{}

	s.config.numSensors = len(conf.Sensors.Coordinates)
	s.config.numGateways = len(conf.Gateways.Coordinates)

	// No seed given, use default.
	if len(conf.Simulation.Seed) == 0 {
		conf.Simulation.Seed = defaultSeed
	}
	// Convert string seed to int64 by hashing.
	seedByte := []byte(conf.Simulation.Seed)
	crc64Table := crc64.MakeTable(0xC96C5795D7870F42)
	s.config.seed = int64(crc64.Checksum(seedByte, crc64Table))

	//s.config.seed = conf.Simulation.Seed
	s.config.duration = conf.Simulation.Duration

	if conf.Simulation.SimulationSpeed == 0 {
		conf.Simulation.SimulationSpeed = defaultSimulationTime
	}
	simtime.SetSimulationSpeed(conf.Simulation.SimulationSpeed)

	if conf.Simulation.Frequency == 0 {
		conf.Simulation.Frequency = defaultFrequency
	}
	s.config.frequency = conf.Simulation.Frequency

	if conf.Simulation.WindowSize == 0 {
		conf.Simulation.WindowSize = defaultWindowSize
	}
	s.config.windowSize = conf.Simulation.WindowSize

	if conf.Simulation.BitCost == 0 {
		conf.Simulation.BitCost = defaultBitCost
	}
	s.config.bitCost = conf.Simulation.BitCost

	// For plotting.
	s.gridJSON.Sensors = conf.Sensors.Coordinates
	s.gridJSON.Gateways = conf.Gateways.Coordinates
}

func (s *Simulator) parseSensorConfig(conf ConfigJSON) {
	s.config.sensors.positions = conf.Sensors.Coordinates

	if conf.Sensors.MaxAttempts == 0 {
		conf.Sensors.MaxAttempts = defaultSensorMaxAttempts
	}
	s.config.sensors.maxAttempts = conf.Sensors.MaxAttempts

	if conf.Sensors.Sensitivity == 0 {
		conf.Sensors.Sensitivity = defaultSensorSensitivity
	}
	s.config.sensors.sensitivity = conf.Sensors.Sensitivity

	if conf.Sensors.TPC == true {
		s.config.sensors.sensorType = sensor.TypeDynamic
	} else {
		s.config.sensors.sensorType = sensor.TypeStatic
	}

	s.config.sensors.sendTPC = conf.Sensors.SendTPC
	if conf.Sensors.SendTPC == true && conf.Sensors.TPC == false {
		fmt.Println("WARNING: send_tpc set to true but tpc set to false. Energy will be wasted by sending the same power level all the time.")
	}

	if conf.Sensors.PayloadSize == 0 {
		conf.Sensors.PayloadSize = defaultSensorPayloadSize
	}
	s.config.sensors.payloadSize = conf.Sensors.PayloadSize

	if conf.Sensors.AddressSize == 0 {
		conf.Sensors.AddressSize = defaultSensorAddressSize
	}
	s.config.sensors.addressSize = conf.Sensors.AddressSize

	if conf.Sensors.CiphertextSize == 0 {
		conf.Sensors.CiphertextSize = defaultSensorCiphertextSize
	}
	s.config.sensors.ciphertextSize = conf.Sensors.CiphertextSize

	if conf.Sensors.SendInterval == 0 {
		conf.Sensors.SendInterval = defaultSensorSendingInterval
	}
	s.config.sensors.sendingInterval = time.Duration(conf.Sensors.SendInterval) * simtime.Minute

	if conf.Sensors.WaitingTime == 0 {
		conf.Sensors.WaitingTime = defaultSensorWaitingTime
	}
	s.config.sensors.waitingTime = time.Duration(conf.Sensors.WaitingTime) * simtime.Minute

	// Dynamic sensor type requires transmission power steps. If none were given, use the default.
	sort.Float64s(conf.Sensors.TransmissionPower)

	if len(conf.Sensors.TransmissionPower) == 1 {
		if conf.Sensors.TransmissionPower[0] == 0 {
			s.config.sensors.transmissionPower = defaultSensorTransmissionSteps[:]
		}
	} else {
		s.config.sensors.transmissionPower = conf.Sensors.TransmissionPower
	}

	if conf.Sensors.DataRate == 0 {
		conf.Sensors.DataRate = defaultSensorDataRate
	}
	s.config.sensors.dataRate = conf.Sensors.DataRate

	if conf.Sensors.StartInterval == 0 {
		if s.config.numSensors > 0 {
			s.config.sensors.startInterval = s.config.sensors.sendingInterval / time.Duration(s.config.numSensors)
		} else {
			conf.Sensors.StartInterval = defaultSensorStartInterval
			s.config.sensors.startInterval = time.Duration(conf.Sensors.StartInterval) * simtime.Second
		}
	} else {
		s.config.sensors.startInterval = time.Duration(conf.Sensors.StartInterval) * simtime.Second
	}
}

func (s *Simulator) parseGatewayConfig(conf ConfigJSON) {
	s.config.gateways.positions = conf.Gateways.Coordinates

	if conf.Gateways.Sensitivity == 0 {
		conf.Gateways.Sensitivity = defaultGatewaySensitivity
	}
	s.config.gateways.sensitivity = conf.Gateways.Sensitivity

	if conf.Gateways.TransmissionPower == 0 {
		conf.Gateways.TransmissionPower = defaultGatewayTransmissionPower
	}
	s.config.gateways.transmissionPower = conf.Gateways.TransmissionPower
}

func (s *Simulator) parseCloudConfig(conf ConfigJSON) {
	if conf.Cloud.WeightDownlink == 0 {
		conf.Cloud.WeightDownlink = defaultWeightDownlink
	}
	s.config.cloud.WeightDownlink = conf.Cloud.WeightDownlink

	if conf.Cloud.WeightUplink == 0 {
		conf.Cloud.WeightUplink = defaultWeightUplink
	}
	s.config.cloud.WeightUplink = conf.Cloud.WeightUplink

	if len(conf.Cloud.SelectionMethod) == 0 {
		fmt.Println("WARNING: no gateway selection type given. Will default to strongest signal strength.")
	}

	s.config.selectionMethod = cloud.SelectionMethodStringToType[conf.Cloud.SelectionMethod]

	if s.config.selectionMethod == cloud.SelectionMethodLowestEnergyConsumption && s.config.sensors.sendTPC == false {
		fmt.Println("WARNING: lowest energy consumption can only be used if the transmission power level is sent from the sensors.")
		fmt.Println("Please check your config file and set send_tpc to true.")
		os.Exit(1)
	}

	if s.config.selectionMethod == cloud.SelectionMethodLowestEnergyConsumption && s.config.sensors.sensorType != sensor.TypeDynamic {
		fmt.Println("WARNING: lowest energy consumption is used without TPC enabled by the sensors. Performance will be lower.")
	}

	if conf.Cloud.MinEnergyImprovement == 0 {
		//conf.Cloud.MinEnergyImprovement = defaultMinimumEnergyImprovement
		conf.Cloud.MinEnergyImprovement = 1.0
	}
	s.config.cloud.MinEnergyImprovement = conf.Cloud.MinEnergyImprovement

	/*
		if conf.Cloud.MinSignalStrengthDifference == 0 {
			conf.Cloud.MinSignalStrengthDifference = defaultMinimumSignalStrengthDifference
		}
	*/
	s.config.cloud.MinSignalStrengthDifference = conf.Cloud.MinSignalStrengthDifference

	/*
		if conf.Cloud.MinReliabilityImprovement == 0 {
			conf.Cloud.MinReliabilityImprovement = defaultMinimumReliabilityImprovement
		}
	*/
	s.config.cloud.MinReliabilityImprovement = conf.Cloud.MinReliabilityImprovement
}

func (s *Simulator) parseConnectionConfig(conf ConfigJSON) {
	if conf.Connections.NormalRange == 0 {
		// Calculate maximum threshold based on sensitivities and transmission power.
		//s.config.threshold = defaultThreshold
		s.config.normalRange = s.calculateThreshold()
	} else {
		s.config.normalRange = conf.Connections.NormalRange
	}

	if conf.Connections.BlockProbability > 100 || conf.Connections.BlockProbability < 0 {
		fmt.Println("WARNING: blocking probability must be between 0 and 100. Defaulting to", defaultBlockProbability)
		conf.Connections.BlockProbability = defaultBlockProbability
	}
	s.config.blockProbability = conf.Connections.BlockProbability

	s.config.symmetricConnections = conf.Connections.SymmetricConnection

	// Iterating over struct fields in golang is rather cumbersome...
	s.connectionDistribution[0] = conf.Connections.Distributions.Standard
	s.connectionDistribution[1] = conf.Connections.Distributions.Dynamic
	s.connectionDistribution[2] = conf.Connections.Distributions.Binary

	// BOOST
	s.boostConnectionDistribution[0] = conf.Connections.BoostDistributions.Standard
	s.boostConnectionDistribution[1] = conf.Connections.BoostDistributions.Dynamic
	s.boostConnectionDistribution[2] = conf.Connections.BoostDistributions.Binary

	// Set default boost range as twice as current range!
	s.config.boostEnabled = conf.Simulation.Boost
	if conf.Simulation.Boost == true {
		if conf.Connections.BoostRange == 0 {
			conf.Connections.BoostRange = 2 * s.config.normalRange
		}
		s.config.boostRange = conf.Connections.BoostRange

		if conf.Connections.BoostCost == 0 {
			conf.Connections.BoostCost = defaultBoostCost
		}
		s.config.boostCost = conf.Connections.BoostCost
	}

	if conf.Connections.DynamicLoss < 0 || conf.Connections.DynamicLoss > 100 {
		fmt.Println("Warning, dynamic connection loss should be between 0 and 100. Defaulting to", defaultDynamicLoss)
		conf.Connections.DynamicLoss = defaultDynamicLoss
	}
	s.config.dynamicLoss = conf.Connections.DynamicLoss

	// Adjust FSPL
	if conf.Connections.AdjustFSPL {

		maxDist := s.calculateThreshold()

		if conf.Connections.NormalRange != 0 {
			// Adjust the distance so that the maximum distance makes the signal strength almost too low
			s.config.fsplAdjustment = (maxDist * 0.95) / s.config.normalRange // Give some leeway? 0.95 of maxdist?
		} else {
			s.config.fsplAdjustment = 1
		}

		if conf.Connections.BoostRange != 0 {
			s.config.fsplAdjustmentBoost = (maxDist * 0.95) / s.config.boostRange
		} else {
			s.config.fsplAdjustmentBoost = 1
		}
	} else {
		s.config.fsplAdjustment = 1
		s.config.fsplAdjustmentBoost = 1
	}
}
