package sensor

import (
	"bytes"
	"encoding/binary"
	"fmt"
	"strings"
	"time"

	"github.com/orjansv/master-code/simulator/connection"
	"github.com/orjansv/master-code/simulator/watchdog"
)

const (
	// TypeStatic is a sensor type that uses constant sending strength.
	TypeStatic int = iota

	// TypeDynamic is a sensor type that dynamically adjusts its sending strength. (TPC)
	TypeDynamic
)

// Message is a message type that is sent between gateways and sensors.
type Message struct {
	SensorID, MessageID, GatewayID int
	SignalStrength                 float64
	Ciphertext                     string
	TransmissionLevel              int
	Address                        int
	PacketSize                     float64
	Boost                          bool
}

// Sensor is a type for simulating a wireless sensor.
type Sensor struct {
	info             Info
	channels         Channels
	connections      map[int]connection.Connection
	boostConnections map[int]connection.Connection
	currentStep      int
	gateway          GatewayInfo
}

// Channels contains the channels used by the sensors.
type Channels struct {
	OutChannels      map[int]chan Message
	OutChannelsBoost map[int]chan Message
	InChannel        chan Message
	WatchdogChannel  chan watchdog.WatchdogMessage
	StopChannel      chan int
}

// Info contains the info the sensor needs to know to work.
type Info struct {
	Sensitivity        float64
	MaxAttempts        int
	TransmissionPowers []float64
	PayloadSize        int
	CiphertextSize     int
	AddressSize        int
	ID                 int
	SensorType         int
	SendingInterval    time.Duration
	WaitingTime        time.Duration
	Address            int
	PacketSize         float64
	Ciphertexts        []string
	BoostModeEnabled   bool
	SendTPC            bool
	VerboseLevel       int
}

// GatewayInfo contains the info that the sensors need to know about the gateways.
type GatewayInfo struct {
	TransmissionPower float64
	Sensitivity       float64
}

// NewSensor creates a new instance of a sensor struct.
func NewSensor(sInfo Info, channels Channels, connections, boostConnections map[int]connection.Connection, gatewayInfo GatewayInfo) Sensor {

	sInfo.Address = sInfo.ID
	sInfo.PacketSize = float64(sInfo.PayloadSize + sInfo.AddressSize)
	if sInfo.SendTPC {
		sInfo.PacketSize++ // 1 byte for holding the transmission power level (could probably be less)
	}

	// Pre-allocate ciphertexts.
	numCiphertexts := sInfo.MaxAttempts
	if sInfo.BoostModeEnabled {
		numCiphertexts *= 2
	}

	sInfo.Ciphertexts = generateCiphertexts(numCiphertexts, sInfo.CiphertextSize)

	s := Sensor{
		info:             sInfo,
		channels:         channels,
		connections:      connections,
		boostConnections: boostConnections,
		gateway:          gatewayInfo,
		currentStep:      len(sInfo.TransmissionPowers) - 1, // Start with highest available power
	}

	return s

}

// Run starts the sensor simulation.
func (s Sensor) Run() {

	msgID := 0
	nextTime := time.Now()
	boostMode := false

	// Send messages periodically.
	for {
		attempts := 0
		nextTime = nextTime.Add(s.info.SendingInterval)

		if nextTime.Sub(time.Now()) < 0 {
			if s.info.VerboseLevel > 0 {
				fmt.Println("WRN Sensor is not executing often enough!", nextTime.Sub(time.Now()), "behind")
			}
		}

		if boostMode {
			// Send message to all gateways in boost mode.
			for gwID, channel := range s.channels.OutChannelsBoost {
				msg := s.CreateMessageBoost(msgID, gwID, attempts)
				select {
				case channel <- msg:
					// Successfully sent message
				case <-s.channels.StopChannel:
					goto Stop
				}
			}
		} else {
			// Send message to all gateways in normal mode.
			for gwID, channel := range s.channels.OutChannels {
				msg := s.CreateMessage(msgID, gwID, attempts)
				select {
				case channel <- msg:
					// Successfully sent message
				case <-s.channels.StopChannel:
					goto Stop
				}
			}
		}

		// Notify Watchdog that a message was broadcast.
		wdMsg := s.CreateWatchdogMessage(false, boostMode)
		select {
		case s.channels.WatchdogChannel <- wdMsg:
			// Successfully sent message to Watchdog.
		case <-s.channels.StopChannel:
			goto Stop
		}

		// This loop runs until a reply is received or the maximum amount of retries is exceeded.
		attempts++
	WaitForReply:
		for {
			// Wait for reply or timeout.
			select {
			case msg := <-s.channels.InChannel:
				if msg.SensorID == s.info.ID && msg.MessageID == msgID {

					var sigStrength float64
					if msg.Boost {
						sigStrength = s.boostConnections[msg.GatewayID].GetSignalStrengthAtSensor(msgID)
					} else {
						sigStrength = s.connections[msg.GatewayID].GetSignalStrengthAtSensor(msgID)
					}

					if sigStrength > s.info.Sensitivity {
						s.UpdateTransmissionPower(sigStrength)

						// If received via non-boost gateway, we can go to normal mode.
						if boostMode {
							if _, ok := s.connections[msg.GatewayID]; ok {
								boostMode = false
							}
						}

						break WaitForReply
					}
				}

			case <-time.After(s.info.WaitingTime):
				// Timeout, increase transmission power (if possible).
				if s.info.SensorType == TypeDynamic {
					if (s.currentStep) < len(s.info.TransmissionPowers)-1 {
						s.currentStep++
					}
				}

				// Try to re-send the message.
				if boostMode {
					// Send message to all gateways in boost mode.
					for gwID, channel := range s.channels.OutChannelsBoost {
						msg := s.CreateMessageBoost(msgID, gwID, attempts)
						select {
						case channel <- msg:
							// Successfully sent message.
						case <-s.channels.StopChannel:
							goto Stop
						}
					}
				} else {
					// Send message to all gateways in normal mode.
					for gwID, channel := range s.channels.OutChannels {
						msg := s.CreateMessage(msgID, gwID, attempts)
						select {
						case channel <- msg:
							// Successfully sent message
						case <-s.channels.StopChannel:
							goto Stop
						}
					}
				}

				// Notify Watchdog that a retry was made.
				wdMsg := s.CreateWatchdogMessage(true, boostMode)
				select {
				case s.channels.WatchdogChannel <- wdMsg:
					// Successfully sent message to Watchdog.
				case <-s.channels.StopChannel:
					goto Stop
				}
				attempts++

			case <-s.channels.StopChannel:
				goto Stop

			}

			// Retries exceeded. Give up sending this message.
			if s.info.BoostModeEnabled {
				if boostMode {
					if attempts >= s.info.MaxAttempts*2 {
						break WaitForReply
					}
				} else if attempts >= s.info.MaxAttempts {
					boostMode = true
				}
			} else if attempts >= s.info.MaxAttempts {
				break WaitForReply
			}
		}

		msgID++

		// Sleep so that it runs every 15 minutes
		time.Sleep(time.Until(nextTime))
	}

Stop:
	// Confirm to cloud that we stopped.
	s.channels.StopChannel <- 0
	return

}

// CreateMessage creates a message for a gateway.
func (s *Sensor) CreateMessage(msgID, gwID, attempt int) Message {

	return Message{
		SensorID:          s.info.ID,
		MessageID:         msgID,
		Ciphertext:        s.info.Ciphertexts[attempt],
		SignalStrength:    s.connections[gwID].GetSignalStrengthAtGateway(msgID, s.info.TransmissionPowers[s.currentStep]),
		TransmissionLevel: s.currentStep,
		Address:           s.info.Address,
		PacketSize:        s.info.PacketSize,
		Boost:             false,
	}
}

// CreateMessageBoost creates a message for a gateway.
func (s *Sensor) CreateMessageBoost(msgID, gwID, attempt int) Message {

	return Message{
		SensorID:          s.info.ID,
		MessageID:         msgID,
		Ciphertext:        s.info.Ciphertexts[attempt],
		SignalStrength:    s.boostConnections[gwID].GetSignalStrengthAtGateway(msgID, s.info.TransmissionPowers[s.currentStep]),
		TransmissionLevel: s.currentStep,
		Address:           s.info.Address,
		PacketSize:        s.info.PacketSize,
		Boost:             true,
	}
}

// CreateWatchdogMessage creates a message for the watchdog.
func (s *Sensor) CreateWatchdogMessage(retry, boostMode bool) watchdog.WatchdogMessage {
	return watchdog.WatchdogMessage{
		SensorID:               s.info.ID,
		NumBytes:               int(s.info.PacketSize),
		TransmissionPowerLevel: s.currentStep,
		Retry: retry,
		Boost: boostMode,
	}
}

// UpdateTransmissionPower calculates the transmission power to use based on the received signal strength.
// This is in essence the TPC (Transmission Power Control)
func (s *Sensor) UpdateTransmissionPower(receivedStrength float64) {
	if s.info.SensorType == TypeDynamic {

		fspl := s.gateway.TransmissionPower - receivedStrength
		powerNeeded := s.gateway.Sensitivity + fspl

		if powerNeeded < s.info.TransmissionPowers[s.currentStep] && s.currentStep > 0 {
			s.currentStep = s.transmissionPowerControl(powerNeeded)
		} else if powerNeeded > s.info.TransmissionPowers[s.currentStep] && s.currentStep < len(s.info.TransmissionPowers)-1 {
			s.currentStep = s.transmissionPowerControl(powerNeeded)
		}

		// If not, the current step is ok.

	} else {
	}
	// Static means transmission power is fixed, so step is also fixed.
}

func (s *Sensor) transmissionPowerControl(new float64) int {

	step := s.currentStep

	// Current transmission power is sufficient, see if we should lower.
	// We currently only lower one step at the time.
	if s.info.TransmissionPowers[step] > new {
		if step > 0 {
			if s.info.TransmissionPowers[step-1] > new {
				step--
			}
		}
	} else {
		// Current transmission power is too low or just enough, see how much we should increase.
		for i := step + 1; i < len(s.info.TransmissionPowers); i++ {
			if s.info.TransmissionPowers[i] >= new {
				return i
			} else if i == len(s.info.TransmissionPowers)-1 {
				return i
			}
		}
	}

	return step
}

func generateCiphertexts(numCiphertexts, ciphertextSize int) []string {
	ciphertexts := make([]string, numCiphertexts)
	ciphertext := "a"
	for i := range ciphertexts {
		ciphertexts[i] = strings.Repeat(ciphertext, ciphertextSize)
		ciphertext = string([]byte(ciphertext)[0] + 1)
	}

	return ciphertexts
}

// This function can be used to increment a ciphertext with letters a-z. It should increment the rightmost first, so you get
// aaa, aab ... ayb, ayc, ... yde, ydf, ... zzy, zzz
func nextCiphertext(s string) string {
	b := []byte(s)
	for i := len(b) - 1; i >= 0; i-- {
		b[i]++
		if b[i] > 122 {
			b[i] = []byte("a")[0]
		} else {
			return string(b)
		}
	}
	return string(b)
}

func addressToByteSlice(address, numBytes int) []byte {

	buf := new(bytes.Buffer)
	err := binary.Write(buf, binary.LittleEndian, int64(address))
	if err != nil {
		panic("Address conversion failed!")
	}

	addrSlice := make([]byte, numBytes)
	for i := 0; i < numBytes; i++ {
		addrSlice[i] = buf.Bytes()[i]
	}

	return addrSlice
}
