package simulator

import (
	"math"

	"github.com/orjansv/master-code/simulator/cloud"
	"github.com/orjansv/master-code/simulator/connection"
	"github.com/orjansv/master-code/simulator/gateway"
	"github.com/orjansv/master-code/simulator/sensor"
	"github.com/orjansv/master-code/simulator/watchdog"
	"github.com/orjansv/master-code/utils"
)

func (s *Simulator) setupCloud() *cloud.Cloud {

	channels := cloud.Channels{
		InChannel:    s.channels.gateway2CloudChan,
		OutChans:     s.channels.cloud2GatewayChans,
		StopChan:     s.stopChannel,
		MessagesChan: s.channels.messagesChan,
	}

	sensorInfo := cloud.SensorInfo{
		TransmissionPowers: s.config.sensors.transmissionPower,
		MaxAttempts:        s.config.sensors.maxAttempts,
		DataRate:           float64(s.config.sensors.dataRate),
		BoostModeEnabled:   s.config.boostEnabled,
		WaitingTime:        s.config.sensors.waitingTime,
	}

	c := cloud.NewCloud(
		channels,
		s.config.selectionMethod,
		sensorInfo,
		s.config.duration,
		s.config.windowSize,
		s.config.cloud,
		s.VerboseLevel,
		s.config.bitCost,
		s.config.boostCost,
		s.config.boostEnabled,
	)
	return &c
}

func (s *Simulator) setupWatchDog() *watchdog.Watchdog {
	w := watchdog.NewWatchDog(
		s.channels.watchdogMessageChan,
		s.channels.watchDogStopChan,
		s.channels.energyStatsChan,
		s.config.sensors.transmissionPower,
		float64(s.config.sensors.dataRate),
		s.config.bitCost,
		s.config.boostCost,
	)
	return &w
}

func (s *Simulator) setupGateways(connectionMatrix, connectionMatrixBoost [][]int) []*gateway.Gateway {

	// Assign IDs to gateways (for now simply incremented)
	gatewayIDs := make([]int, s.config.numGateways)
	for i := range gatewayIDs {
		gatewayIDs[i] = i
	}

	gateways := make([]*gateway.Gateway, s.config.numGateways)

	for index, id := range gatewayIDs {
		// Create map with channels in range of the current gateway
		sensorChannels := make(map[int]chan sensor.Message)

		for i := range connectionMatrix {
			if connectionMatrix[i][index] == 1 {
				sensorChannels[i] = s.channels.gateway2SensorChans[i]
			}

			if s.config.boostEnabled {
				if connectionMatrixBoost[i][index] == 1 {
					sensorChannels[i] = s.channels.gateway2SensorChans[i]
				}
			}
		}

		gatewayChannels := gateway.Channels{
			Gateway2CloudChan:  s.channels.gateway2CloudChan,
			Cloud2GatewayChan:  s.channels.cloud2GatewayChans[index],
			Gateway2SensorChan: sensorChannels,
			Sensor2GatewayChan: s.channels.sensor2GatewayChans[index],
			StopChan:           s.channels.gatewayStopChans[index],
		}

		g := gateway.NewGateway(id, s.config.gateways.sensitivity, gatewayChannels)

		gateways[index] = &g
	}

	return gateways

}

func (s *Simulator) setupSensors(connectionMatrix, connectionMatrixBoost [][]int, distanceMatrix [][]float64, connectionTypes, boostConnectionTypes []int) []*sensor.Sensor {

	sensorIDs := make([]int, s.config.numSensors)
	for i := range sensorIDs {
		sensorIDs[i] = i
	}

	sensors := make([]*sensor.Sensor, s.config.numSensors)

	blockThreshold := math.Min(s.config.sensors.sensitivity, s.config.gateways.sensitivity)

	connTypeCounter := 0
	connTypeCounterBoost := 0
	connTypeCounterBoost2 := 0
	for index, id := range sensorIDs {

		numChannels := utils.Sum(connectionMatrix[index][:])
		channels := make(map[int]chan sensor.Message, numChannels)
		connections := make(map[int]connection.Connection, numChannels)

		var numBoost int
		var boostChannels map[int]chan sensor.Message
		var boostConnections map[int]connection.Connection

		if s.config.boostEnabled {
			numBoost = utils.Sum(connectionMatrixBoost[index][:])
			boostChannels = make(map[int]chan sensor.Message, numBoost)
			boostConnections = make(map[int]connection.Connection, numBoost)
		}

		for i := 0; i < s.config.numGateways; i++ {
			if connectionMatrix[index][i] == 1 {

				// There is a connection, so we need a channel.
				channels[i] = s.channels.sensor2GatewayChans[i]

				// We also need a connection type for determining signal strength at the sensor.
				// It is an array with indexes corresponding to the channels array.
				connections[i] = connection.NewConnection(
					distanceMatrix[index][i],
					s.config.frequency,
					s.config.gateways.transmissionPower,
					blockThreshold,
					s.config.fsplAdjustment,
					connectionTypes[connTypeCounter],
					s.config.blockProbability,
					s.config.dynamicLoss,
					int64(id+i)+s.config.seed, // to ensure different seed for different connections, but to make it the same throughout simulations.
					s.config.symmetricConnections,
				)

				connTypeCounter++
			}

			// Add new boost connections + normal connections (as both are reached in boost mode)
			if s.config.boostEnabled {
				if connectionMatrixBoost[index][i] == 1 {
					// Use existing channels (same as for normal mode)
					boostChannels[i] = s.channels.sensor2GatewayChans[i]

					// Use same connection type as previously, if there is one
					var connType int
					if connectionMatrix[index][i] == 1 {
						connType = connectionTypes[connTypeCounterBoost]
						connTypeCounterBoost++
					} else {
						connType = boostConnectionTypes[connTypeCounterBoost2]
						connTypeCounterBoost2++
					}

					boostConnections[i] = connection.NewConnection(
						distanceMatrix[index][i],
						s.config.frequency,
						s.config.gateways.transmissionPower,
						blockThreshold,
						s.config.fsplAdjustmentBoost,
						connType,
						s.config.blockProbability,
						s.config.dynamicLoss,
						int64(id+i*2)+s.config.seed,
						s.config.symmetricConnections,
					)
				}
			}
		}

		sensorInfo := sensor.Info{
			ID:                 id,
			SensorType:         s.config.sensors.sensorType,
			Sensitivity:        s.config.sensors.sensitivity,
			MaxAttempts:        s.config.sensors.maxAttempts,
			TransmissionPowers: s.config.sensors.transmissionPower,
			PayloadSize:        s.config.sensors.payloadSize,
			AddressSize:        s.config.sensors.addressSize,
			SendingInterval:    s.config.sensors.sendingInterval,
			WaitingTime:        s.config.sensors.waitingTime,
			CiphertextSize:     s.config.sensors.ciphertextSize,
			BoostModeEnabled:   s.config.boostEnabled,
			SendTPC:            s.config.sensors.sendTPC,
			VerboseLevel:       s.VerboseLevel,
		}

		sensorChannels := sensor.Channels{
			OutChannels:      channels,
			OutChannelsBoost: boostChannels,
			InChannel:        s.channels.gateway2SensorChans[index],
			WatchdogChannel:  s.channels.watchdogMessageChan,
			StopChannel:      s.channels.sensorStopChans[index],
		}

		gatewayInfo := sensor.GatewayInfo{
			TransmissionPower: s.config.gateways.transmissionPower,
			Sensitivity:       s.config.gateways.sensitivity,
		}

		s := sensor.NewSensor(
			sensorInfo,
			sensorChannels,
			connections,
			boostConnections,
			gatewayInfo,
		)

		sensors[index] = &s
	}

	return sensors

}
